"""Main entry point for optom_tools."""

from optom_tools.prescription import Prescription
from optom_tools.utils import log
from optom_tools.visual_acuity import VisualAcuity

__all__ = ["Prescription", "VisualAcuity", "log"]


def main():
    """Entry function."""
    pass


if __name__ == "__main__":
    main()
